@@ -1,13 +1,51 @@
-#!/usr/bin/env python
-# -*- coding: utf-8 -*-
+from dataclasses import dataclass
+from typing import Dict, Any, Optional
 
 from lighttree.interactive import Obj
+from elasticsearch import Elasticsearch
 
+from pandagg import Mappings, MappingsDict
 from pandagg.interactive.mappings import IMappings
 from pandagg.search import Search
 
 
-def discover(using, index="*"):
+@dataclass
+class Index:
+    name: str
+    settings: Dict[str, Any]
+    mappings: MappingsDict
+    aliases: Any
+    client: Optional[Elasticsearch] = None
+
+    @property
+    def imappings(self) -> IMappings:
+        return IMappings(
+            Mappings(**self.mappings), client=self.client, index=[self.name]
+        )
+
+    def search(
+        self, nested_autocorrect: bool = True, repr_auto_execute: bool = True
+    ) -> Search:
+        return Search(
+            using=self.client,
+            mappings=self.mappings,
+            index=self.name,
+            nested_autocorrect=nested_autocorrect,
+            repr_auto_execute=repr_auto_execute,
+        )
+
+    def __str__(self) -> str:
+        return self.__repr__()
+
+    def __repr__(self) -> str:
+        return "<Index '%s'>" % self.name
+
+
+class Indices(Obj):
+    _COERCE_ATTR = True
+
+
+def discover(using: Elasticsearch, index: str = "*") -> Indices:
     """
     :param using: Elasticsearch client
     :param index: Comma-separated list or wildcard expression of index names used to limit the request.
@@ -22,36 +60,3 @@ def discover(using, index="*"):
             aliases=index_detail["aliases"],
         )
     return indices
-
-
-# until Proper Index class is written
-
-
-class Index(object):
-    def __init__(self, name, settings, mappings, aliases, client=None):
-        super(Index, self).__init__()
-        self.client = client
-        self.name = name
-        self.settings = settings
-        self._mappings = mappings
-        self.mappings = IMappings(mappings, client=client, index=name)
-        self.aliases = aliases
-
-    def search(self, nested_autocorrect=True, repr_auto_execute=True):
-        return Search(
-            using=self.client,
-            mappings=self._mappings,
-            index=self.name,
-            nested_autocorrect=nested_autocorrect,
-            repr_auto_execute=repr_auto_execute,
-        )
-
-    def __str__(self):
-        return self.__repr__()
-
-    def __repr__(self):
-        return "<Index '%s'>" % self.name
-
-
-class Indices(Obj):
-    _COERCE_ATTR = True
