@@ -1,5 +1,3 @@
-#!/usr/bin/env python
-# -*- coding: utf-8 -*-
 """Not implemented aggregations include:
 - children agg
 - geo-distance
@@ -12,6 +10,8 @@
 
 from pandagg.node.types import NUMERIC_TYPES
 from pandagg.node.aggs.abstract import MultipleBucketAgg, UniqueBucketAgg
+from pandagg.types import Meta, BucketKey, QueryClauseDict, RangeDict, BucketDict
+from typing import Any, Optional, Dict, Union, List
 
 
 class Global(UniqueBucketAgg):
@@ -19,10 +19,10 @@ class Global(UniqueBucketAgg):
     KEY = "global"
     VALUE_ATTRS = ["doc_count"]
 
-    def __init__(self, meta=None):
+    def __init__(self, meta: Meta = None):
         super(Global, self).__init__(agg_body={}, meta=meta)
 
-    def get_filter(self, key):
+    def get_filter(self, key: BucketKey) -> Optional[QueryClauseDict]:
         return None
 
 
@@ -31,7 +31,9 @@ class Filter(UniqueBucketAgg):
     KEY = "filter"
     VALUE_ATTRS = ["doc_count"]
 
-    def __init__(self, filter=None, meta=None, **body):
+    def __init__(
+        self, filter: Optional[Dict[str, Any]] = None, meta: Meta = None, **body: Any
+    ):
         if (filter is not None) != (not body):
             raise ValueError(
                 'Filter aggregation requires exactly one of "filter" or "body"'
@@ -43,12 +45,12 @@ class Filter(UniqueBucketAgg):
         self.filter = filter_
         super(Filter, self).__init__(meta=meta, **filter_)
 
-    def get_filter(self, key):
+    def get_filter(self, key: BucketKey) -> Optional[QueryClauseDict]:
         return self.filter
 
 
 class MatchAll(Filter):
-    def __init__(self, meta=None, **body):
+    def __init__(self, meta: Meta = None, **body: Any):
         super(MatchAll, self).__init__(filter={"match_all": {}}, meta=meta, **body)
 
 
@@ -58,11 +60,11 @@ class Nested(UniqueBucketAgg):
     VALUE_ATTRS = ["doc_count"]
     WHITELISTED_MAPPING_TYPES = ["nested"]
 
-    def __init__(self, path, meta=None, **body):
+    def __init__(self, path: str, meta: Meta = None, **body: Any):
         self.path = path
         super(Nested, self).__init__(path=path, meta=meta, **body)
 
-    def get_filter(self, key):
+    def get_filter(self, key: BucketKey) -> Optional[QueryClauseDict]:
         return None
 
 
@@ -72,26 +74,26 @@ class ReverseNested(UniqueBucketAgg):
     VALUE_ATTRS = ["doc_count"]
     WHITELISTED_MAPPING_TYPES = ["nested"]
 
-    def __init__(self, path=None, meta=None, **body):
+    def __init__(self, path: Optional[str] = None, meta: Meta = None, **body: Any):
         self.path = path
         body_kwargs = dict(body)
         if path:
             body_kwargs["path"] = path
         super(ReverseNested, self).__init__(meta=meta, **body_kwargs)
 
-    def get_filter(self, key):
+    def get_filter(self, key: BucketKey) -> Optional[QueryClauseDict]:
         return None
 
 
 class Missing(UniqueBucketAgg):
     KEY = "missing"
     VALUE_ATTRS = ["doc_count"]
-    BLACKLISTED_MAPPING_TYPES = []
 
-    def __init__(self, field, meta=None, **body):
+    def __init__(self, field: str, meta: Meta = None, **body: Any):
+        self.field: str = field
         super(UniqueBucketAgg, self).__init__(field=field, meta=meta, **body)
 
-    def get_filter(self, key):
+    def get_filter(self, key: BucketKey) -> Optional[QueryClauseDict]:
         return {"bool": {"must_not": {"exists": {"field": self.field}}}}
 
 
@@ -100,12 +102,18 @@ class Terms(MultipleBucketAgg):
 
     KEY = "terms"
     VALUE_ATTRS = ["doc_count", "doc_count_error_upper_bound", "sum_other_doc_count"]
-    BLACKLISTED_MAPPING_TYPES = []
 
-    def __init__(self, field, missing=None, size=None, meta=None, **body):
-        self.field = field
-        self.missing = missing
-        self.size = size
+    def __init__(
+        self,
+        field: str,
+        missing: Optional[Union[int, str]] = None,
+        size: Optional[int] = None,
+        meta: Optional[Meta] = None,
+        **body: Any
+    ) -> None:
+        self.field: str = field
+        self.missing: Optional[Union[int, str]] = missing
+        self.size: Optional[int] = size
 
         body_kwargs = dict(body)
         if missing is not None:
@@ -115,7 +123,7 @@ class Terms(MultipleBucketAgg):
 
         super(Terms, self).__init__(field=field, meta=meta, **body_kwargs)
 
-    def get_filter(self, key):
+    def get_filter(self, key: BucketKey) -> Optional[QueryClauseDict]:
         """Provide filter to get documents belonging to document of given key."""
         if key == "missing":
             return {"bool": {"must_not": {"exists": {"field": self.field}}}}
@@ -130,11 +138,17 @@ class Filters(MultipleBucketAgg):
     IMPLICIT_KEYED = True
 
     def __init__(
-        self, filters, other_bucket=False, other_bucket_key=None, meta=None, **body
-    ):
-        self.filters = filters
-        self.other_bucket = other_bucket
-        self.other_bucket_key = other_bucket_key
+        self,
+        filters: Dict[str, QueryClauseDict],
+        other_bucket: bool = False,
+        other_bucket_key: Optional[str] = None,
+        meta: Optional[Meta] = None,
+        **body: Any
+    ) -> None:
+        self.filters: Dict[str, QueryClauseDict] = filters
+        self.other_bucket: bool = other_bucket
+        self.other_bucket_key: Optional[str] = other_bucket_key
+
         body_kwargs = dict(body)
         if other_bucket:
             body_kwargs["other_bucket"] = other_bucket
@@ -143,7 +157,7 @@ class Filters(MultipleBucketAgg):
 
         super(Filters, self).__init__(filters=filters, meta=meta, **body_kwargs)
 
-    def get_filter(self, key):
+    def get_filter(self, key: BucketKey) -> Optional[QueryClauseDict]:
         """Provide filter to get documents belonging to document of given key."""
         if key in self.filters.keys():
             return self.filters[key]
@@ -163,14 +177,16 @@ class Histogram(MultipleBucketAgg):
     VALUE_ATTRS = ["doc_count"]
     WHITELISTED_MAPPING_TYPES = NUMERIC_TYPES
 
-    def __init__(self, field, interval, meta=None, **body):
-        self.field = field
-        self.interval = interval
+    def __init__(
+        self, field: str, interval: int, meta: Optional[Meta] = None, **body: Any
+    ) -> None:
+        self.field: str = field
+        self.interval: int = interval
         super(Histogram, self).__init__(
             field=field, interval=interval, meta=meta, **body
         )
 
-    def get_filter(self, key):
+    def get_filter(self, key: BucketKey) -> Optional[QueryClauseDict]:
         try:
             key = float(key)
         except (TypeError, ValueError):
@@ -189,21 +205,22 @@ class DateHistogram(MultipleBucketAgg):
 
     def __init__(
         self,
-        field,
-        interval=None,
-        calendar_interval=None,
-        fixed_interval=None,
-        meta=None,
-        keyed=False,
-        key_as_string=True,
-        **body
-    ):
+        field: str,
+        interval: str = None,
+        calendar_interval: str = None,
+        fixed_interval: str = None,
+        meta: Meta = None,
+        keyed: bool = False,
+        key_as_string: bool = True,
+        **body: Any
+    ) -> None:
         """Date Histogram aggregation.
         Note: interval is deprecated from 7.2 in favor of calendar_interval and fixed interval
         :param keyed: defines returned buckets format: if True, as dict.
         :param key_as_string: if True extracted key of bucket will be the formatted date (applicable if keyed=False)
         """
-        self.field = field
+        self.field: str = field
+
         if not (interval or fixed_interval or calendar_interval):
             raise ValueError(
                 'One of "interval", "calendar_interval" or "fixed_interval" must be provided.'
@@ -214,6 +231,7 @@ class DateHistogram(MultipleBucketAgg):
             body["calendar_interval"] = calendar_interval
         if fixed_interval:
             body["fixed_interval"] = fixed_interval
+
         self.interval = interval or calendar_interval or fixed_interval
         super(DateHistogram, self).__init__(
             field=field,
@@ -223,7 +241,7 @@ class DateHistogram(MultipleBucketAgg):
             **body
         )
 
-    def get_filter(self, key):
+    def get_filter(self, key: BucketKey) -> Optional[QueryClauseDict]:
         # https://www.elastic.co/guide/en/elasticsearch/reference/current/common-options.html#date-math
         return {
             "range": {self.field: {"gte": key, "lt": "%s||+%s" % (key, self.interval)}}
@@ -234,11 +252,22 @@ class Range(MultipleBucketAgg):
     KEY = "range"
     VALUE_ATTRS = ["doc_count"]
     WHITELISTED_MAPPING_TYPES = NUMERIC_TYPES
-    KEY_SEP = "-"
+    KEY_SEP: str = "-"
+
+    bucket_key_suffix: Optional[str]
+
+    def __init__(
+        self,
+        field: str,
+        ranges: List[RangeDict],
+        keyed: bool = False,
+        meta: Optional[Meta] = None,
+        **body: Any
+    ) -> None:
+        self.field: str = field
+        self.ranges: List[RangeDict] = ranges
+        self.bucket_key_suffix: Optional[str]
 
-    def __init__(self, field, ranges, keyed=False, meta=None, **body):
-        self.field = field
-        self.ranges = ranges
         body_kwargs = dict(body)
         if keyed:
             self.bucket_key_suffix = "_as_string"
@@ -249,18 +278,18 @@ class Range(MultipleBucketAgg):
         )
 
     @property
-    def from_key(self):
+    def from_key(self) -> str:
         if self.bucket_key_suffix:
             return "from%s" % self.bucket_key_suffix
         return "from"
 
     @property
-    def to_key(self):
+    def to_key(self) -> str:
         if self.bucket_key_suffix:
             return "to%s" % self.bucket_key_suffix
         return "to"
 
-    def _extract_bucket_key(self, bucket):
+    def _extract_bucket_key(self, bucket: BucketDict) -> BucketKey:
         if self.from_key in bucket:
             key = "%s%s" % (bucket[self.from_key], self.KEY_SEP)
         else:
@@ -271,7 +300,7 @@ class Range(MultipleBucketAgg):
             key += "*"
         return key
 
-    def get_filter(self, key):
+    def get_filter(self, key: str) -> Dict[str, Any]:
         from_, to_ = key.split(self.KEY_SEP)
         inner = {}
         if from_ != "*":
@@ -286,15 +315,15 @@ class DateRange(Range):
     VALUE_ATTRS = ["doc_count"]
     WHITELISTED_MAPPING_TYPES = ["date"]
     # cannot use range '-' separator since some keys contain it
-    KEY_SEP = "::"
+    KEY_SEP: str = "::"
+
+    def __init__(
+        self,
+        field: str,
+        key_as_string: bool = True,
+        meta: Optional[Meta] = None,
+        **body: Any
+    ) -> None:
+        self.key_as_string: bool = key_as_string
 
-    def __init__(self, field, key_as_string=True, meta=None, **body):
-        self.key_as_string = key_as_string
         super(DateRange, self).__init__(field=field, keyed=True, meta=meta, **body)
-
-
-class Composite(MultipleBucketAgg):
-    KEY = "composite"
-
-    def get_filter(self, key):
-        raise NotImplementedError()
