@@ -1,3 +1,6 @@
+from typing import List, Optional, Any
+
+from pandagg.types import QueryName
 from pandagg.node.query.abstract import QueryClause
 
 
@@ -16,11 +19,11 @@ class CompoundClause(QueryClause):
 
     """
 
-    _default_operator = None
-    # key -> multiple / single
-    _parent_params = {}
+    KEY: str
+    _default_operator: str = ""
+    _parent_params: List[str] = []
 
-    def __init__(self, _name=None, **body):
+    def __init__(self, _name: Optional[QueryName] = None, **body: Any) -> None:
         b = body.copy()
         children = {}
         for param in self._parent_params:
@@ -39,7 +42,7 @@ class Bool(CompoundClause):
     """
 
     _default_operator = "must"
-    _parent_params = {"should": True, "must": True, "must_not": True, "filter": True}
+    _parent_params = ["should", "must", "must_not", "filter"]
     KEY = "bool"
 
 
