@@ -1,4 +1,7 @@
+from typing import Any, Optional, Tuple
+
 from .abstract import KeyFieldQueryClause, AbstractSingleFieldQueryClause
+from pandagg.types import DistanceType, ValidationMethod
 
 
 class GeoBoundingBox(KeyFieldQueryClause):
@@ -8,15 +11,21 @@ class GeoBoundingBox(KeyFieldQueryClause):
 class GeoDistance(AbstractSingleFieldQueryClause):
     KEY = "geo_distance"
 
-    def __init__(self, distance, **body):
+    def __init__(self, distance: str, **body: Any) -> None:
         # pop all allowed args to find out which keyword is used as field
-        _name = body.pop("_name", None)
-        distance_type = body.pop("distance_type", None)
-        validation_method = body.pop("validation_method", None)
+        _name: Optional[str] = body.pop("_name", None)
+        distance_type: Optional[DistanceType] = body.pop("distance_type", None)
+        validation_method: Optional[ValidationMethod] = body.pop(
+            "validation_method", None
+        )
+
         if len(body) != 1:
             raise ValueError("Wrong declaration: %s" % body)
+
         field, location = self.expand__to_dot(body).popitem()
-        self.field = field
+
+        self.field: str = field
+
         b = {field: location}
         if distance_type is not None:
             b["distance_type"] = distance_type
@@ -26,7 +35,7 @@ class GeoDistance(AbstractSingleFieldQueryClause):
             _name=_name, field=field, distance=distance, **b
         )
 
-    def line_repr(self, depth, **kwargs):
+    def line_repr(self, depth: int, **kwargs: Any) -> Tuple[str, str]:
         return self.KEY, "field=%s" % self.field
 
 
