@@ -1,9 +1,5 @@
-#!/usr/bin/env python
-# -*- coding: utf-8 -*-
-
-
 # adapted from https://github.com/elastic/elasticsearch-dsl-py/blob/master/elasticsearch_dsl/utils.py#L162
-from six import add_metaclass
+from typing import Dict, Tuple, Any
 
 
 class DslMeta(type):
@@ -12,20 +8,27 @@ class DslMeta(type):
     for given DslBase subclass (== all the query types for the Query subclass
     of DslBase).
 
-    It then uses the information from that registry (as well as `name` and
-    `deserializer` attributes from the base class) to construct any subclass based
-    on it's name.
+    Types will be: 'agg', 'query', 'field'
+
+    Each of those types will hold a `_classes` dictionary pointing to all classes of same type.
     """
 
-    _types = {}
+    # registry for types
+    _types: Dict[str, "DslMeta"] = {}
+    # per type, registry for classes (not initialized here)
+    _classes: Dict[str, "DslMeta"]
 
-    def __init__(cls, name, bases, attrs):
+    # types keys
+    _type_name: str = ""
+    # classes keys
+    KEY: str = ""
+
+    def __init__(cls, name: str, bases: Tuple, attrs: Dict) -> None:
         super(DslMeta, cls).__init__(name, bases, attrs)
-        # skip for DSLMixin
-        if not hasattr(cls, "_type_name") or cls._type_name is None:
+        if not cls._type_name:
+            # skip for DSLMixin
             return
-
-        if not hasattr(cls, "KEY") or cls.KEY is None:
+        if not cls.KEY:
             # abstract base class, register its shortcut
             cls._types[cls._type_name] = cls
             # and create a registry for subclasses
@@ -36,13 +39,12 @@ class DslMeta(type):
             cls._classes[cls.KEY] = cls
 
 
-@add_metaclass(DslMeta)
-class DSLMixin(object):
+class DSLMixin(metaclass=DslMeta):
     """Base class for all DSL objects - queries, filters, aggregations etc. Wraps
     a dictionary representing the object's json."""
 
     @classmethod
-    def _get_dsl_class(cls, name):
+    def get_dsl_class(cls, name: str) -> DslMeta:
         try:
             return cls._classes[name]
         except KeyError:
@@ -51,14 +53,14 @@ class DSLMixin(object):
             )
 
     @staticmethod
-    def _get_dsl_type(name):
+    def get_dsl_type(name: str) -> DslMeta:
         try:
             return DslMeta._types[name]
         except KeyError:
             raise ValueError("DSL type %s does not exist." % name)
 
 
-def ordered(obj):
+def ordered(obj: Any) -> Any:
     if isinstance(obj, dict):
         return sorted((k, ordered(v)) for k, v in obj.items())
     if isinstance(obj, list):
@@ -66,12 +68,12 @@ def ordered(obj):
     return obj
 
 
-def equal_queries(d1, d2):
+def equal_queries(d1: Any, d2: Any) -> bool:
     """Compares if two queries are equivalent (do not consider nested list orders)."""
     return ordered(d1) == ordered(d2)
 
 
-def equal_search(s1, s2):
+def equal_search(s1: Any, s2: Any) -> bool:
     if not isinstance(s1, dict) or not isinstance(s2, dict):
         raise ValueError("not a search")
     s1 = s1.copy()
