@@ -1,8 +1,3 @@
-#!/usr/bin/env python
-# -*- coding: utf-8 -*-
-
-from __future__ import unicode_literals
-
 from unittest import TestCase
 
 from elasticsearch import Elasticsearch
@@ -43,4 +38,4 @@ class WrapperTestCase(TestCase):
         self.assertEqual(report_index.name, "classification_report_one")
 
         # ensure mappings presence
-        self.assertIsInstance(report_index.mappings, IMappings)
+        self.assertIsInstance(report_index.imappings, IMappings)
