@@ -1,12 +1,9 @@
-#!/usr/bin/env python
-# -*- coding: utf-8 -*-
-
 from tests import PandaggTestCase
 import pandas as pd
 
 from pandagg.search import Search
 from pandagg.tree.response import AggsResponseTree
-from pandagg.response import Response, Hits, Hit, Aggregations
+from pandagg.response import SearchResponse, Hits, Hit, Aggregations
 from pandagg.tree.aggs import Aggs
 
 import tests.testing_samples.data_sample as sample
@@ -19,14 +16,12 @@ class ResponseTestCase(PandaggTestCase):
         h = Hit(
             {
                 "_index": "my_index_01",
-                "_type": "_doc",
                 "_id": "1",
                 "_score": 1.0,
                 "_source": {"field_23": 1},
             }
         )
         self.assertEqual(h._index, "my_index_01")
-        self.assertEqual(h._type, "_doc")
         self.assertEqual(h._id, "1")
         self.assertEqual(h._score, 1.0)
         self.assertEqual(h._source, {"field_23": 1})
@@ -90,7 +85,7 @@ class ResponseTestCase(PandaggTestCase):
         self.assertEqual(non_expanded_df.index.tolist(), ["1", "2"])
 
     def test_response(self):
-        r = Response(
+        r = SearchResponse(
             {
                 "took": 42,
                 "timed_out": False,
@@ -140,7 +135,7 @@ class AggregationsResponseTestCase(PandaggTestCase):
     def test_parse_as_tree(self, *_):
         my_agg = Aggs(sample.EXPECTED_AGG_QUERY, mappings=MAPPINGS)
         response = Aggregations(
-            data=sample.ES_AGG_RESPONSE, search=Search().aggs(my_agg)
+            data=sample.ES_AGG_RESPONSE, _search=Search().aggs(my_agg)
         ).to_tree()
         self.assertIsInstance(response, AggsResponseTree)
         self.assertEqual(response.__str__(), sample.EXPECTED_RESPONSE_TREE_REPR)
@@ -148,7 +143,7 @@ class AggregationsResponseTestCase(PandaggTestCase):
     def test_normalize_buckets(self):
         my_agg = Aggs(sample.EXPECTED_AGG_QUERY, mappings=MAPPINGS)
         response = Aggregations(
-            data=sample.ES_AGG_RESPONSE, search=Search().aggs(my_agg)
+            data=sample.ES_AGG_RESPONSE, _search=Search().aggs(my_agg)
         ).to_normalized()
         self.assertEqual(
             ordered(response), ordered(sample.EXPECTED_NORMALIZED_RESPONSE)
@@ -158,7 +153,7 @@ class AggregationsResponseTestCase(PandaggTestCase):
         # with single agg at root
         my_agg = Aggs(sample.EXPECTED_AGG_QUERY, mappings=MAPPINGS)
         index_names, index_values = Aggregations(
-            data=sample.ES_AGG_RESPONSE, search=Search().aggs(my_agg)
+            data=sample.ES_AGG_RESPONSE, _search=Search().aggs(my_agg)
         ).to_tabular(index_orient=True, grouped_by="global_metrics.field.name")
 
         self.assertEqual(
@@ -192,7 +187,7 @@ class AggregationsResponseTestCase(PandaggTestCase):
 
         # index_orient = False
         index_names, index_values = Aggregations(
-            data=sample.ES_AGG_RESPONSE, search=Search().aggs(my_agg)
+            data=sample.ES_AGG_RESPONSE, _search=Search().aggs(my_agg)
         ).to_tabular(index_orient=False, grouped_by="global_metrics.field.name")
 
         self.assertEqual(
@@ -255,7 +250,7 @@ class AggregationsResponseTestCase(PandaggTestCase):
             "avg_f1_score": {"value": 0.815},
         }
         index_names, index_values = Aggregations(
-            data=raw_response, search=Search().aggs(my_agg)
+            data=raw_response, _search=Search().aggs(my_agg)
         ).to_tabular(index_orient=True, expand_sep=" || ")
 
         self.assertEqual(index_names, [])
@@ -272,7 +267,7 @@ class AggregationsResponseTestCase(PandaggTestCase):
 
         # with specified grouped_by
         index_names, index_values = Aggregations(
-            data=raw_response, search=Search().aggs(my_agg)
+            data=raw_response, _search=Search().aggs(my_agg)
         ).to_tabular(grouped_by="classification_type")
         self.assertEqual(index_names, ["classification_type"])
         self.assertEqual(
@@ -283,7 +278,7 @@ class AggregationsResponseTestCase(PandaggTestCase):
     def test_parse_as_dataframe(self):
         my_agg = Aggs(sample.EXPECTED_AGG_QUERY, mappings=MAPPINGS)
         df = Aggregations(
-            data=sample.ES_AGG_RESPONSE, search=Search().aggs(my_agg)
+            data=sample.ES_AGG_RESPONSE, _search=Search().aggs(my_agg)
         ).to_dataframe(grouped_by="global_metrics.field.name")
         self.assertIsInstance(df, pd.DataFrame)
         self.assertEqual(
@@ -322,7 +317,7 @@ class AggregationsResponseTestCase(PandaggTestCase):
     def test_grouping_agg(self):
         my_agg = Aggs(sample.EXPECTED_AGG_QUERY, mappings=MAPPINGS)
         agg_response = Aggregations(
-            data=sample.ES_AGG_RESPONSE, search=Search().aggs(my_agg)
+            data=sample.ES_AGG_RESPONSE, _search=Search().aggs(my_agg)
         )
 
         # none provided
