@@ -2,10 +2,10 @@ from collections import OrderedDict
 from unittest import TestCase
 from mock import patch
 
-from pandagg.tree.aggs.aggs import Aggs
+from pandagg.tree.aggs import Aggs
 from pandagg.tree.response import AggsResponseTree
 
-from tests.testing_samples.mapping_example import MAPPING
+from tests.testing_samples.mapping_example import MAPPINGS
 import tests.testing_samples.data_sample as sample
 
 
@@ -13,7 +13,7 @@ class ResponseTestCase(TestCase):
     @patch("uuid.uuid4")
     def test_response_tree(self, uuid_mock):
         uuid_mock.side_effect = range(1000)
-        my_agg = Aggs(sample.EXPECTED_AGG_QUERY, mapping=MAPPING)
+        my_agg = Aggs(sample.EXPECTED_AGG_QUERY, mappings=MAPPINGS)
         response_tree = AggsResponseTree(aggs=my_agg).parse(sample.ES_AGG_RESPONSE)
         self.assertEqual(response_tree.__str__(), sample.EXPECTED_RESPONSE_TREE_REPR)
         self.assertEqual(len(response_tree.list()), 15)
